//	gcc -m64 -shared accessmath_features.c -o accessmath_features.so

#include <stdio.h>
#include <math.h>

#define TypeLogPolar 	0
#define TypeEquidistant 1

void print_normalized_cc(unsigned char *cc_image, int size, const char *black, const char *white){
	printf("\n");
	
    int x, y;
	for (y = 0; y < size; y++){
		for (x = 0; x < size; x++){
			printf((int)cc_image[y * size + x] > 128 ? white : black);
		}
		printf("\n");
	}	
}

/*
	Computes AVG X, AVG Y, VAR X, VAR Y, COV X-Y
*/
void compute_stats_CPU(unsigned char* cc_in, int cc_size, double* features_out, int feature_offset){
	//features.append(n_ratio)             
    //features += self.getStats()
	double half = cc_size / 2.0;

	//For each pixel...
	int offset;
	int count = 0;
	double sx, sy;
	double avg_sx = 0.0;
	double avg_sy = 0.0;
	int x, y;
	
	for (y = 0; y < cc_size; y++){
		offset = (y  * cc_size);
		sy = (y - half) / half;

		for (x = 0; x < cc_size; x++){
			sx = (x - half) / half;

			//If the pixel is white....
			if (cc_in[offset] > 0){
				//Center of mass...
				avg_sx += sx;
				avg_sy += sy;
				count++;
			}

			offset++;
		}
	}

	//Get averages...
	avg_sx /= (double)count;
	avg_sy /= (double)count;

	//And now get the variance and covariance...
	double var_x = 0.0;
    double var_y = 0.0;
    double cov_xy = 0.0;
	offset = 0;
	for (y = 0; y < cc_size; y++){
		offset = (y  * cc_size);
		sy = (y - half) / half;

		for (x = 0; x < cc_size; x++){
			sx = (x - half) / half;

			//If the pixel is white....
			if (cc_in[offset] > 128){
				var_x += (sx - avg_sx) * (sx - avg_sx);
                var_y += (sy - avg_sy) * (sy - avg_sy);
				cov_xy += (sx - avg_sx) * (sy - avg_sy);        
			}

			offset++;
		}
	}

	var_x /= (double)count;
	var_y /= (double)count;
	cov_xy /= (double)count;

	features_out[feature_offset + 0] = avg_sx;
	features_out[feature_offset + 1] = avg_sy;
	features_out[feature_offset + 2] = var_x;
	features_out[feature_offset + 3] = var_y;
	features_out[feature_offset + 4] = cov_xy;
}

/*
	Compute crossing features in the given direction
*/
void compute_crossings_CPU_direction(unsigned char* cc_in, int cc_size, double* features_out, int feature_offset, 
									int n_crossings, int horizontal){

	double step = cc_size / (n_crossings + 1.0);
	int i, k;
	
	for (i = 1; i <= n_crossings; i++){
		int pos = (int)(i * step);

		// the crossing is treated in terms of boolean intervals...
		int interval_start = -1;
		int interval_end = -1;
		int mid_point;
		int current_count = 0;
		double current_min = 1.1;
		double current_max = -1.1;
		double scaled_mid_point;
		int crossing_found = 0;
		k = 0;
		
		while (k < cc_size){
			//Assume no crossing is found...
			crossing_found = 0;

			if (horizontal){
				//horizontal -> y fixed and x moves
				if (cc_in[pos * cc_size + k] > 128){
					//New white interval found....
					crossing_found = 1;
					interval_start = k;
					//...move until end of current interval...
					while (k < cc_size && cc_in[pos * cc_size + k] > 128){
						k++;
					}
					interval_end = k - 1;
				} else {
					k++;
				}
			} else {
				//vertical -> x fixed and y moves
				if (cc_in[k * cc_size + pos] > 128){
					//New white interval found...
					crossing_found = 1;
					interval_start = k;
					//...move until end of current interval...
					while (k < cc_size && cc_in[k * cc_size + pos] > 128){
						k++;
					}
					interval_end = k - 1;
				} else {
					k++;
				}
			}

			if (crossing_found){
				//Find mid point...
				mid_point = (interval_start + interval_end) / 2;

				//Scale mid point...
				scaled_mid_point = (mid_point / (double)(cc_size - 1)) * 2.0 - 1.0; 
				
				//Add to count...
				current_count++;

				//Mark if first....
				if (scaled_mid_point < current_min){
					current_min = scaled_mid_point;
				}				
				
				//Mark as last....
				current_max = scaled_mid_point;
			}
		}

		//Scale the count...
		if (current_count > 10){
			current_count = 10;
		}		
		
		//Add crossing feature values to output vector....
		//...counts...
		features_out[feature_offset + (i - 1)] = ((current_count / 10.0) * 6.0 - 3.0);
		features_out[feature_offset + n_crossings + (i - 1)] = current_min;
		features_out[feature_offset + n_crossings * 2 + (i - 1)] = current_max;				
	}
}

/*
	Compute crossing features
*/
void compute_crossings_CPU(unsigned char* cc_in, int cc_size, double* features_out, int feature_offset, 
						   int n_hor_crossings, int n_ver_crossings){
	//...Horizontal crossings...
	if (n_hor_crossings > 0){
		compute_crossings_CPU_direction(cc_in, cc_size, features_out, feature_offset, n_hor_crossings, 1);
	}
	
	//...Vertical crossings...
	if (n_ver_crossings > 0){
		compute_crossings_CPU_direction(cc_in, cc_size, features_out, feature_offset + n_hor_crossings * 3, n_ver_crossings, 0);	
	}
}

/*
	Compute 2D histogram features
*/
void compute_2d_histogram_CPU(unsigned char* cc_in, int cc_size, double* features_out, int feature_offset, 
							  int hist_size, double histogram_scale){

	int start_y = 0;
	int start_x = 0;
	int height, width;
	int total_count = 0;
	int local_count = 0;
	int offset;
	int block_offset = feature_offset;
	int row, col, x, y;
	for (row = 0; row < hist_size; row++){
		height = (int)(cc_size / (float)hist_size) +  (row < (cc_size % hist_size) ? 1 : 0);

		start_x = 0;
		for (col = 0; col < hist_size; col++){
			width = (int)(cc_size / (float)hist_size) +  (col < (cc_size % hist_size) ? 1 : 0);

			//Now count
			local_count = 0;
			for (y = start_y; y < start_y + height; y++){
				offset = (y * cc_size) + start_x;
				for (x = start_x; x < start_x + width; x++){
					if (cc_in[offset] > 128){
						local_count++;
					}

					offset++;	
				}
			}

			//Assign...
			features_out[block_offset] = local_count;
			total_count += local_count;

			//Move to next block...
			block_offset++;
			start_x += width;			
		}
		//Move to next block...
		start_y += height;
	}

	//now, normalize....
	//cout << "[";
	block_offset = feature_offset;
	for (row = 0; row < hist_size; row++){
		//cout << "{";
		for (col = 0; col < hist_size; col++){
			features_out[block_offset] = ((features_out[block_offset] / (float)total_count) - 0.5f) * histogram_scale;
			
			block_offset++;
		}
		//cout << "}";
	}	
	//cout << "]" << endl;
}


void get_cc_features(unsigned char *cc_image, int size, int use_aspect, int use_stats, 
					int n_hor_crossings, int n_ver_crossings, int hist_size, 
					double hist_scale, double* out_features){
	//First, global features...
	//Assumes that aspect ratio is pre-computed on Python side
	int feature_offset = (use_aspect ? 1 : 0);
	
	//Check stats...
	if (use_stats){
		compute_stats_CPU(cc_image, size, out_features, feature_offset);
		feature_offset += 5;
	}
	
	//Add Crossings...
	if (n_hor_crossings > 0 || n_ver_crossings > 0){
		compute_crossings_CPU(cc_image, size, out_features, feature_offset, n_hor_crossings, n_ver_crossings);
		feature_offset += (n_hor_crossings + n_ver_crossings) * 3;
	}
	
	//Add 2D histogram...
	if (hist_size > 0){
		compute_2d_histogram_CPU(cc_image, size, out_features, feature_offset, hist_size, hist_scale);
		feature_offset += hist_size * hist_size;
	}
}

void compute_shape_contexts_CPU(unsigned char *image, int width, int height, 
								int radial_bins, int polar_bins, int radial_type, 
								int* key_points, int n_key_points, float* out_features)
{
	
	float start_angle = -M_PI / polar_bins;
	int kp, px, py;
	
	//for idx, kp in enumerate(keypoints):
	for (kp = 0; kp < n_key_points; kp++){
		int kp_x = key_points[kp * 3 + 0];
		int kp_y = key_points[kp * 3 + 1];
		int kp_radius = key_points[kp * 3 + 2];
		
		//printf("Point (%d, %d, %d) \n", kp_x, kp_y, kp_radius);
		
		int squared_kp_radius = kp_radius * kp_radius;
		int squared_radius;
		float first_rel_radius = (1.0f / pow(2.0f, radial_bins - 1));
		
		int total_pixels = 0;
		int radial_bin;
		
		//for px in range(kp.x - kp.radius, kp.x + kp.radius + 1):
		for (px = kp_x - kp_radius; px < kp_x + kp_radius + 1; px++){
			//for py in range(kp.y - kp.radius, kp.y + kp.radius + 1):
			for (py = kp_y - kp_radius; py < kp_y + kp_radius + 1; py++){
				//now, validate if pixel in image range...
				//then, check if black pixel...
				
				//...get radius...
				//radius = math.sqrt(math.pow(kp.x - px, 2) + math.pow(kp.y - py, 2))
				squared_radius = (kp_x - px) * (kp_x - px) + (kp_y - py) * (kp_y - py);
				
				if (px >= 0 && py >= 0 && px <= width - 1 && py <= height - 1 &&					
					image[py * width + px] == 0 && squared_radius <= squared_kp_radius){
					
					float radius = sqrt((float)squared_radius);
					float rel_radius = radius / (float)kp_radius;
					
					//...compute bins....
					//...radial...
					if (radial_type == TypeLogPolar){
						//use log-polar system, log of radius
						if (rel_radius < first_rel_radius){
							//special case... first bin...
							radial_bin = 0;
						}else{
							//the others...
							radial_bin = (int)fmin(radial_bins + log2(rel_radius), radial_bins - 1);
						}
					}else{
						if (radial_type == TypeEquidistant){
							//equidistant bins...
							radial_bin = (int)fmin(rel_radius * radial_bins, radial_bins - 1);
						}else{
							radial_bin = 0;
						}
					}
				
					//...angular....
					float angle = atan2((float)(py - kp_y), (float)(px - kp_x));
					float rel_angle = angle - start_angle;
					while (rel_angle < 0){
						rel_angle += (float)(M_PI * 2.0);
					}
					while (rel_angle >= M_PI * 2.0){
						rel_angle -= (float)(M_PI * 2.0);
					}

					int polar_bin = (int)((rel_angle / (M_PI * 2.0)) * polar_bins);
				
					out_features[(kp * radial_bins + radial_bin) * polar_bins + polar_bin] ++;
					total_pixels ++;
				}
			}
		}
		
		
		//now, normalize....
		if (total_pixels > 0){
			int feat_idx;
			int kp_offset = kp * radial_bins * polar_bins;
			
			for (feat_idx = 0; feat_idx < radial_bins * polar_bins; feat_idx++){
				out_features[kp_offset + feat_idx] /= total_pixels;
			}
		}
	}
	
}