
# ==============================================================================
#	AccessMath V 0.1 - Lecture Video Summarization
#	Copyright (C) 2017  Kenny Davila, Richard Zanibbi
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ===============================================================================

import cv2
import sys
import time

from AccessMath.data.meta_data_DB import MetaDataDB
from AccessMath.annotation.keyframe_annotation import KeyFrameAnnotation
from AccessMath.annotation.unique_cc_group import UniqueCCGroup
from AccessMath.util.time_helper import TimeHelper
from AccessMath.evaluation.evaluator import Evaluator

import numpy as np
import matplotlib.pyplot as plt

def main():
    if len(sys.argv) < 3:
        print("Usage: python " + sys.argv[0] + " database lecture")
        print("Where")
        print("\tdatabase\t= Database metadata file")
        print("\tlecture\t= Lecture to export")
        print("\tbaseline\t= [Optional] Baseline name")
        return

    database, lecture = MetaDataDB.load_database_lecture(sys.argv[1], sys.argv[2])
    if database is None:
        return

    if len(sys.argv) >= 4:
        base_line_prefix = sys.argv[3] + "_"
    else:
        base_line_prefix = ""


    # load output from pipeline ...
    lecture_sufix = str(lecture.id) + ".dat"

    summary_prefix = database.output_summaries + "/" + base_line_prefix + database.name + "_" + lecture.title.lower()
    annotation_prefix = database.output_annotations + "/" + database.name + "_" + lecture.title.lower()

    annot_filename = annotation_prefix + "/segments.xml"
    annot_cc_groups_filename = annotation_prefix + "/unique_ccs.xml"
    annot_image_prefix = annotation_prefix + "/keyframes/"
    annot_binary_prefix = annotation_prefix + "/binary/"

    print("... loading data ...")
    start_loading = time.time()

    # ideal summary ...
    annot_keyframes, annot_segments = KeyFrameAnnotation.LoadExportedKeyframes(annot_filename, annot_image_prefix, True)
    for keyframe in annot_keyframes:
        keyframe.binary_image = cv2.imread(annot_binary_prefix + str(keyframe.idx) + ".png")
        keyframe.update_binary_cc(False)

    annot_keyframes = KeyFrameAnnotation.CombineKeyframesPerSegment(annot_keyframes, annot_segments, False)

    annot_cc_group, annot_unique_groups = UniqueCCGroup.GroupsFromXML(annot_keyframes, annot_cc_groups_filename)

    # provided summary ...
    summ_filename = summary_prefix + "/segments.xml"
    summ_image_prefix = summary_prefix + "/keyframes/"
    summ_keyframes, summ_segments = KeyFrameAnnotation.LoadExportedKeyframes(summ_filename, summ_image_prefix, True, False, True)
    for keyframe in summ_keyframes:
        # keyframe.binary_image = keyframe.raw_image.copy()
        keyframe.update_binary_cc(False)

    summ_keyframes = KeyFrameAnnotation.CombineKeyframesPerSegment(summ_keyframes, summ_segments, False)

    print("... data loaded!")

    print("... computing metrics ...")
    start_evaluation = time.time()

    all_metrics, ranges = Evaluator.compute_summary_metrics(annot_segments, annot_keyframes, annot_unique_groups,
                                                            annot_cc_group, summ_segments, summ_keyframes, False)

    print("")
    print("")
    print("Evaluation Summary")

    for range in ranges:
        print("")
        print("*---------------*")
        print("")
        Evaluator.print_summary_recall_metrics(all_metrics[range], range)
        print("")
        Evaluator.print_summary_precision_metrics(all_metrics[range], range)

    note = """
        Note: Ground truth recall is computed globally, in terms of global unique connected components, but
            summaries are considered in terms of all raw connected components. This means precision will count
            multiple instances of the same element as separate CC if it appears in mutliple frames.
    """
    print(note)

    print("")
    Evaluator.print_compact_CC_metrics(all_metrics["all"], "all")


    end_evaluation = time.time()

    print("Total Summary Keyframes: " + str(len(summ_keyframes)))
    print("Time Loading:    " + TimeHelper.secondsToStr(start_evaluation - start_loading))
    print("Time Evaluation: " + TimeHelper.secondsToStr(end_evaluation - start_evaluation))

if __name__ == '__main__':
    main()