
# ==============================================================================
#	AccessMath V 0.1 - Lecture Video Summarization
#	Copyright (C) 2017  Kenny Davila, Richard Zanibbi
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ===============================================================================

#============================================================================
# Preprocessing Model for ST3D indexing - V - Ideal
#
# Kenny Davila
# - Created:  November 15, 2016
# - Modified: April 2, 2017
#   Adapted for v1.1 of pipeline where BG removal is applied before stability analysis
#
#============================================================================

import sys
import cv2
import numpy as np

from AccessMath.preprocessing.config.parameters import Parameters
from AccessMath.preprocessing.content.helper import Helper
from AccessMath.preprocessing.content.roi_background_remover import ROIBackgroundRemover
from AccessMath.preprocessing.user_interface.console_ui_process import ConsoleUIProcess

from AccessMath.annotation.keyframe_annotation import KeyFrameAnnotation




def process_input(process, input_data):
    frame_times, frame_indices, compressed_frames = input_data

    print("... loading GT data ...")
    annotation_prefix = process.database.output_annotations + "/" + process.database.name + "_" + process.current_lecture.title.lower()

    annot_filename = annotation_prefix + "/segments.xml"
    annot_image_prefix = annotation_prefix + "/keyframes/"
    annot_binary_prefix = annotation_prefix + "/binary/"

    # ideal summary ...
    annot_keyframes, annot_segments = KeyFrameAnnotation.LoadExportedKeyframes(annot_filename, annot_image_prefix, True,
                                                                               False)
    for keyframe in annot_keyframes:
        keyframe.binary_image = cv2.imread(annot_binary_prefix + str(keyframe.idx) + ".png")
        keyframe.update_binary_cc()

    print("... GT data loaded! ")

    # parameters
    # 1) decompress all images ...
    print("Decompressing input...")
    all_binary = Helper.decompress_binary_images(compressed_frames)

    print("Computing frames...")
    # identify first the corresponding key-frames
    current_kf_idx = 0
    compressed_output = []

    per_gt_kf_binaries = []
    ROI_masks = []
    for keyframe in annot_keyframes:
        per_gt_kf_binaries.append([])
        ROI_masks.append(np.logical_not(keyframe.object_mask).astype(np.uint8) * 255)

    for frame_idx in range(len(all_binary)):
        abs_frame_idx = frame_indices[frame_idx]
        while current_kf_idx < len(annot_keyframes) and abs_frame_idx >= annot_keyframes[current_kf_idx].idx:
            current_kf_idx += 1

        if current_kf_idx < len(annot_keyframes):
            per_gt_kf_binaries[current_kf_idx].append(all_binary[frame_idx])
        else:
            per_gt_kf_binaries[current_kf_idx - 1].append(all_binary[frame_idx])

    for kf_idx, keyframe in enumerate(annot_keyframes):
        current_bins = ROIBackgroundRemover.remove_background(per_gt_kf_binaries[kf_idx], ROI_masks[kf_idx],
                                                              Parameters.ROIBGRem_min_overlap)
        compressed_output += current_bins

    return frame_times, frame_indices, compressed_output

def main():
    #usage check
    if not ConsoleUIProcess.usage_check(sys.argv):
        return

    process = ConsoleUIProcess(sys.argv[1], sys.argv[2:], Parameters.Output_Binarize, Parameters.Output_GT_BG_Removal)

    # (this is really difficult to obtain, too noisy, only works for the general model, not the refined)
    if not process.initialize():
       return

    process.start_input_processing(process_input)

    print("Finished")

if __name__ == "__main__":
    main()